# -*- coding: UTF-8 -*-
"""

Main program: including configuration, data reading, logging, plotting, model training and prediction
"""

import pandas as pd
import numpy as np
import os
import sys
import time
import logging
from logging.handlers import RotatingFileHandler
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
import sklearn
import math
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
frame = "pytorch_multilstm"  # "keras", "pytorch", "tensorflow"
# if frame == "pytorch_attention":
#     from model.model_pytorch_multiatt import train, predict
if frame == "pytorch_multiatt":
    from model.model_pytorch_multiatt import train, predict
elif frame == "pytorch_use":
    from model.model_pytorch_use import train, predict
elif frame == "pytorch_bilstm":
    from model.model_pytorch_bilstm import train, predict
elif frame == "pytorch_multitest":
    from model.model_pytorch_multitest import train, predict
elif frame == "pytorch_attusenew":
    from model.model_pytorch_attusenew import train, predict
elif frame == "pytorch_multiattention":
    from model.model_pytorch_multiattention import train, predict
elif frame == "pytorch_multia":
    from model.model_pytorch_multia import train, predict
elif frame == "pytorch_multilstm":
    from model.model_pytorch_multiattentionlstm import train, predict
elif frame == "pytorch_attuse":
    from model.model_pytorch_attuse import train, predict
elif frame == "pytorch_transformer":
    from model.model_pytorch_transformer import train, predict
elif frame == "pytorch_trans":
    from model.model_transformer import train, predict
elif frame == "pytorch_att":
    from model.model_pytorch_att import train, predict
elif frame == "pytorch_svm":
    from model.model_svm import train, predict
elif frame == "keras":
    from model.model_keras import train, predict
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = '3'
elif frame == "tensorflow":
    from model.model_tensorflow import train, predict
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = '3'    # There will be a lot of tf warning under tf and keras, but it does not affect the training
# else:
#     raise Exception("Wrong frame seletion")

class Config:
    # Data parameters
    feature_columns = list(range(2,32))     # The column to be used as the feature is calculated from 0 by the original data, and can also be set with a list such as [2,4,6,8].
    label_columns = [2]                  # The columns to be predicted are calculated from 0, e.g., the fourth and fifth columns are predicted at the same time.

    label_in_feature_index = (lambda x,y: [x.index(i) for i in y])(feature_columns, label_columns)  # Because FEATURES don't necessarily start at 0

    predict_day = 1             # Forecast for the next few days

    # Network Parameters
    input_size = len(feature_columns)
    output_size = len(label_columns)

    # attention_size=8            # attention层大小
    hidden_size = 64        # The size of the hidden layer of the LSTM, which is also the output size
    lstm_layers = 1          # Number of stacking layers of LSTM
    dropout_rate = 0.001     # dropout probability
    time_step = 10              # This parameter is very important, is to set the number of days before the data to predict, is also the time step number of LSTM, please ensure that the amount of training data is greater than it
    heads = 3
    batch_size = 64
    learning_rate = 0.001
    epoch = 300                 # How many times the entire training set was trained, without considering the premise of early stopping

    # Training parameters multiatt
    do_train = True
    do_predict = True
    add_train = False           # Whether to load existing model parameters for incremental training
    shuffle_train_data = False   # Whether to do shuffle on training data
    use_cuda = False            # 是Whether to use GPU training
    #transformer parameters
    n_layers=6
    num_heads=3
    model_dim=512
    forward_dim=2048
    dropout=0.1
    # valid_lens=0.5

    train_data_rate = 0.8      # Training data as a percentage of the overall data, test data is 1-train_data_rate
    valid_data_rate = 0.25      # Validation data to training data ratio, validation set used in the training process, in order to do model and parameter selection

    patience = 50                # How many epochs to train and stop when the validation set is not boosted
    random_seed = 42            # Random seeds, guaranteed reproducible

    do_continue_train = False    # The final_state of the previous training is used as the init_state of the next training, only for RNN type models, currently only pytorch is supported.
    continue_flag = ""           # However, the actual results are not good, probably because: it can only be trained with batch_size = 1
    if do_continue_train:
        shuffle_train_data = False
        batch_size = 1
        continue_flag = "continue_"

    # Training mode
    debug_mode = False  # Debug mode is to run through the code, the pursuit of fast
    debug_num = 500  # Debug only with debug_num bar data

    # Frame parameters
    used_frame = frame  # The selected deep learning framework, different framework models save different suffixes
    model_postfix = {"pytorch_multiatt": ".pth","pytorch_use": ".pth","pytorch_svm": ".pth","pytorch_multiattention": ".pth",
                     "pytorch_attuse": ".pth","pytorch_transformer": ".pth", "pytorch_att": ".pth","keras": ".h5",
                     "pytorch_multilstm": ".pth","pytorch_multia": ".pth","tensorflow": ".ckpt",
                     "pytorch_trans": ".pth","pytorch_multitest": ".pth", "pytorch_bilstm": ".pth","pytorch_attusenew": ".pth"}
    model_name = "model_" + continue_flag + used_frame + model_postfix[used_frame]

    # Path parameters
    train_data_path = "./data/600104data30.xlsx"
    model_save_path = "./checkpoint/" + used_frame + "/"
    figure_save_path = "./figure/"
    log_save_path = "./600104/multi/"
    exper_save_path = "./exper/"
    do_log_print_to_screen = True
    do_log_save_to_file = True                  # Whether to log the config and training process to the log
    do_figure_save = False
    do_train_visualized = False          # For training loss visualization, pytorch uses visdom, tf uses tensorboardX, which can actually be generalized, keras does not
    if not os.path.exists(model_save_path):
        os.makedirs(model_save_path)    # makedirs Recursive directory creation
    if not os.path.exists(figure_save_path):
        os.mkdir(figure_save_path)
    if do_train and (do_log_save_to_file or do_train_visualized):
        cur_time = time.strftime("%Y_%m_%d_%H_%M_%S", time.localtime())
        log_save_path = log_save_path + cur_time + '_' + used_frame+'epoch'+ str(epoch) +'batch'+str(batch_size) + "/"
        os.makedirs(log_save_path)


class Data:
    def __init__(self, config):
        self.config = config
        self.data, self.data_column_name = self.read_data()

        self.data_num = self.data.shape[0]
        self.train_num = int(self.data_num * self.config.train_data_rate)

        self.mean = np.mean(self.data, axis=0)              # Means and variances of the data
        self.std = np.std(self.data, axis=0)
        self.norm_data = (self.data - self.mean)/self.std   # Normalization, de-quantization

        self.start_num_in_test = 0      # Data from previous days in the test set will be deleted because it is not enough for a time_step

    def read_data(self):                # Read initial data
        if self.config.debug_mode:
            init_data = pd.read_excel(self.config.train_data_path, nrows=self.config.debug_num,
                                    usecols=self.config.feature_columns)
        else:
            init_data = pd.read_excel(self.config.train_data_path, usecols=self.config.feature_columns)
        print("init_data",init_data)
        return init_data.values, init_data.columns.tolist()     # .columns.tolist() is to get the column name

    def get_train_and_valid_data(self):
        print("trainnum",self.train_num)
        feature_data = self.norm_data[:self.train_num]
        label_data = self.norm_data[self.config.predict_day : self.config.predict_day + self.train_num,
                                    self.config.label_in_feature_index]    # Use the data delayed by a few days as label

        if not self.config.do_continue_train:
            # In non-continuous training mode, each time_step row of data will be used as a sample, and the two samples will be staggered by one row, for example: rows 1-20, rows 2-21 。。。。
            train_x = [feature_data[i:i+self.config.time_step] for i in range(self.train_num-self.config.time_step)]
            train_y = [label_data[i:i+self.config.time_step] for i in range(self.train_num-self.config.time_step)]
        else:

            train_x = [feature_data[start_index + i*self.config.time_step : start_index + (i+1)*self.config.time_step]
                       for start_index in range(self.config.time_step)
                       for i in range((self.train_num - start_index) // self.config.time_step)]
            train_y = [label_data[start_index + i*self.config.time_step : start_index + (i+1)*self.config.time_step]
                       for start_index in range(self.config.time_step)
                       for i in range((self.train_num - start_index) // self.config.time_step)]

        train_x, train_y = np.array(train_x), np.array(train_y)

        train_x, valid_x, train_y, valid_y = train_test_split(train_x, train_y, test_size=self.config.valid_data_rate,
                                                              random_state=self.config.random_seed,
                                                              shuffle=self.config.shuffle_train_data)   # 划分训练和验证集，并打乱
        # print(train_x)
        return train_x, valid_x, train_y, valid_y

    def get_test_data(self, return_label_data=False):
        feature_data = self.norm_data[self.train_num:]
        sample_interval = min(feature_data.shape[0], self.config.time_step)     # 防止time_step大于测试集数量
        self.start_num_in_test = feature_data.shape[0] % sample_interval  # 这些天的数据不够一个sample_interval
        time_step_size = feature_data.shape[0] // sample_interval

        # In the test data, each time_step row of data will be used as a sample, and the two samples are staggered time_step rows
        # For example: rows 1-20, rows 21-40. Go to the end of the data.
        test_x = [feature_data[self.start_num_in_test+i*sample_interval : self.start_num_in_test+(i+1)*sample_interval]
                   for i in range(time_step_size)]
        print("testx",np.shape(test_x))
        if return_label_data:       # The test set in the actual application is without label data
            label_data = self.norm_data[self.train_num + self.start_num_in_test:, self.config.label_in_feature_index]
            return np.array(test_x), label_data
        return np.array(test_x)

def load_logger(config):
    logger = logging.getLogger()
    logger.setLevel(level=logging.DEBUG)

    # StreamHandler
    if config.do_log_print_to_screen:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setLevel(level=logging.INFO)
        formatter = logging.Formatter(datefmt='%Y/%m/%d %H:%M:%S',
                                      fmt='[ %(asctime)s ] %(message)s')
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    # FileHandler
    if config.do_log_save_to_file:
        file_handler = RotatingFileHandler(config.log_save_path + "out.log", maxBytes=1024000, backupCount=5)
        file_handler.setLevel(level=logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        # Log the config information to the log file as well
        config_dict = {}
        for key in dir(config):
            if not key.startswith("_"):
                config_dict[key] = getattr(config, key)
        config_str = str(config_dict)
        config_list = config_str[1:-1].split(", '")
        config_save_str = "\nConfig:\n" + "\n'".join(config_list)
        logger.info(config_save_str)

    return logger

def draw(config: Config, origin_data: Data, logger, predict_norm_data: np.ndarray):
    label_data = origin_data.data[origin_data.train_num + origin_data.start_num_in_test : ,
                                            config.label_in_feature_index]
    predict_data = predict_norm_data * origin_data.std[config.label_in_feature_index] + \
                   origin_data.mean[config.label_in_feature_index]   # Data reduction by saved means and variances
    assert label_data.shape[0]==predict_data.shape[0], "The element number in origin and predicted data is different"
    pre_vo=[]
    label_vo=[]
    for i in range(len(predict_data)):
        if i == 0:
            pre_vo.append(predict_data[i]-3550.8767)
        elif i > 0:
            x=predict_data[i]-predict_data[i-1]
            pre_vo.append(x)
    # print(pre_vo)
    # print("y_test:" , y_test)
    for i in range(len(label_data)):
        if i == 0:
            label_vo.append(label_data[i]-3550.8767)
        elif i>0:
            label_vo.append(label_data[i]-label_data[i-1])
    TP = 0
    FP = 0
    TN = 0
    FN = 0
    for i in range(len(pre_vo)):
        if pre_vo[i] > 0:
            if label_vo[i] > 0:
                TP += 1
            elif label_vo[i] < 0:
                FP += 1
        elif pre_vo[i] < 0:
            if label_vo[i] > 0:
                FN += 1
            elif label_vo[i] < 0:
                TN += 1
    precision = TP / (TP + FP)
    accuracy = (TP + TN) / (len(pre_vo))
    recall = TP / (TP + FN)
    fscore = 2 * precision * recall / (precision + recall)
    mse = sklearn.metrics.mean_squared_error(label_data, predict_data)
    # calculate RMSE
    rmse = math.sqrt(sklearn.metrics.mean_squared_error(label_data, predict_data))
    # calculate MAE
    mae = sklearn.metrics.mean_absolute_error(label_data,predict_data)
    mape = sklearn.metrics.mean_absolute_percentage_error(label_data,predict_data)
    # calculate R square
    r_square = sklearn.metrics.r2_score(label_data, predict_data)
    print('mse: %.6f' % mse)
    print('rmse: %.6f' % rmse)
    print('mae: %.6f' % mae)
    print('mape: %.6f' % mape)
    print('R_square: %.6f' % r_square)
    print('precision: %.6f' % precision)
    print('accuracy: %.6f' % accuracy)
    print('recall: %.6f' % recall)
    print('fscore: %.6f' % fscore)
    label_name = [origin_data.data_column_name[i] for i in config.label_in_feature_index]
    label_column_num = len(config.label_columns)


    loss = np.mean((label_data[config.predict_day:] - predict_data[:-config.predict_day] ) ** 2, axis=0)
    loss_norm = loss/(origin_data.std[config.label_in_feature_index] ** 2)
    logger.info("The mean squared error of stock {} is ".format(label_name) + str(loss_norm))
    zhix=[]
    label_X = range(origin_data.data_num - origin_data.train_num - origin_data.start_num_in_test)
    predict_X = [ x + config.predict_day for x in label_X]
    # print("labelx",label_data)
    if not sys.platform.startswith('linux'):    # If you have a desktop Linux, such as Ubuntu, you can remove this line
        for i in range(label_column_num):

            plt.figure(i+1)                     # Predictive data mapping
            plt.plot(label_X, label_data[:, i], label='label')
            print("labelx",label_data[:, i])
            plt.plot(predict_X, predict_data[:, i], label='predict')
            plt.title("Predict stock {} price with {}".format(label_name[i], config.used_frame))
            logger.info("The predicted stock {} for the next {} day(s) is: ".format(label_name[i], config.predict_day) +
                  str(np.squeeze(predict_data[-config.predict_day:, i])))
            if config.do_figure_save:
                plt.savefig(config.figure_save_path+"{}predict_{}_with_{}.png".format(config.continue_flag, label_name[i], config.used_frame))
        plt.show()
    for i in range(len(label_data)):
        zhi1 = {}
        if i == 0:
            zhi1.update({"predict": predict_data[i], "label": label_data[i],"mape":mape,
                         "mse":mse,"rmse":rmse,"mae":mae,"precision":precision,"accuracy":accuracy,
                         "recall":recall,"fscore":fscore,"batch_size":config.batch_size ,
                         "epoch":config.epoch,"lstm_layer":config.lstm_layers,"dropout_rate":config.dropout_rate,
                         "hidden_size":config.hidden_size,"learning_rate":config.learning_rate,
                         "head":config.heads})
        else:
            zhi1.update({"predict": predict_data[i], "label": label_data[i]})
        zhix.append(zhi1)
    # zhi1.update({})
    # zhix.append(zhi1)
    pf = pd.DataFrame(zhix)
    print(pf)
    exper_save_path=config.log_save_path
    pf.to_excel(exper_save_path + str(mape)+'.xlsx')
def draw2(config: Config, origin_data: Data, logger):
    label_data = origin_data.data[origin_data.train_num + origin_data.start_num_in_test : ,
                                            config.label_in_feature_index]
    # predict_data = predict_norm_data * origin_data.std[config.label_in_feature_index] + \
    #                origin_data.mean[config.label_in_feature_index]   # Data reduction by saved means and variances
    # assert label_data.shape[0]==predict_data.shape[0], "The element number in origin and predicted data is different"
    label_name = [origin_data.data_column_name[i] for i in config.label_in_feature_index]
    label_column_num = len(config.label_columns)

    label_X = range(origin_data.data_num - origin_data.train_num - origin_data.start_num_in_test)

    # predict_X = [ x + config.predict_day for x in label_X]
    if not sys.platform.startswith('linux'):    # If you have a desktop Linux, such as Ubuntu, you can remove this line
        for i in range(label_column_num):
            plt.figure(i+1)                     # Predictive data mapping
            plt.plot(label_X, label_data[:, i], label='label')
            # plt.plot(predict_X, predict_data[:, i], label='predict')
            plt.title("Predict stock {} price with {}".format(label_name[i], config.used_frame))
            # logger.info("The predicted stock {} for the next {} day(s) is: ".format(label_name[i], config.predict_day) +
            #       str(np.squeeze(predict_data[-config.predict_day:, i])))
            if config.do_figure_save:
                plt.savefig(config.figure_save_path+"{}predict_{}_with_{}.png".format(config.continue_flag, label_name[i], config.used_frame))

        plt.show()
def main(config):
    logger = load_logger(config)
    try:
        np.random.seed(config.random_seed)  # Set random seeds to ensure reproducibility
        data_gainer = Data(config)

        if config.do_train:
            train_X, valid_X, train_Y, valid_Y = data_gainer.get_train_and_valid_data()
            # pred_result=
            # draw2(config, data_gainer, logger)
            pred_result = train(config, logger, [train_X, train_Y, valid_X, valid_Y])
            # print("pred_resultp",pred_result.shape)
            # draw(config, data_gainer, logger, pred_result)
        if config.do_predict:
            test_X, test_Y = data_gainer.get_test_data(return_label_data=True)
            pred_result = predict(config, test_X)       # The output here is the unreduced normalized prediction data
            # print("pred_result")
            # print(pred_result.shape)
            draw(config, data_gainer, logger, pred_result)

    except Exception:
        logger.error("Run Error", exc_info=True)


if __name__=="__main__":
    import argparse
    # argparsent
    parser = argparse.ArgumentParser()
    # parser.add_argument("-t", "--do_train", default=False, type=bool, help="whether to train")
    # parser.add_argument("-p", "--do_predict", default=True, type=bool, help="whether to train")
    # parser.add_argument("-b", "--batch_size", default=64, type=int, help="batch size")
    # parser.add_argument("-e", "--epoch", default=20, type=int, help="epochs num")
    args = parser.parse_args()

    con = Config()
    for key in dir(args):               # dir(args) function gets all the properties of args
        if not key.startswith("_"):     # Remove args' own attributes, such as __name__, etc.
            setattr(con, key, getattr(args, key))   # Assigning property values to Config

    main(con)
